@@ -3,7 +3,6 @@
 import copy
 import random
 from dataclasses import dataclass
-
 from typing import List, Optional
 
 from .mechanisms import Mechanism, MechanismRoundResult, MechanismState
@@ -45,7 +44,7 @@ class GAMechanism(Mechanism):
         self.ranks = {}
 
     def extra_state(self):
-        return GAState(dominant_outcomes=self.dominant_outcomes)
+        return dict(dominant_outcomes=self.dominant_outcomes)
 
     def generate(self, n: int) -> List[Outcome]:
         return self.random_outcomes(n)
