@@ -58,8 +58,12 @@ class MechanismRoundResult:
     error_details: str = ""
     """Error message"""
     waiting: bool = False
-    """whether to consider that the round is still running and call the the round method again without increasing
+    """whether to consider that the round is still running and call the round method again without increasing
     the step number"""
+    exceptions: Optional[Dict[str, List[str]]] = None
+    """A mapping from negotiator ID to a list of exceptions raised by that negotiator in this round"""
+    times: Optional[Dict[str, float]] = None
+    """A mapping from negotiator ID to the time it consumed during this round"""
 
 
 # noinspection PyAttributeOutsideInit
@@ -230,7 +234,11 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
         )
         self.ami._mechanism = self
 
-        self._history = []
+        self._history: List[MechanismState] = []
+        self._stats: Dict[str, Any] = dict()
+        self._stats["round_times"] = list()
+        self._stats["times"] = defaultdict(float)
+        self._stats["exceptions"] = defaultdict(list)
         # if self.ami.issues is not None:
         #     self.ami.issues = tuple(self.ami.issues)
         # if self.ami.outcomes is not None:
@@ -351,12 +359,12 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
         )
 
     @property
-    def time(self) -> Optional[float]:
-        """Elapsed time since mechanism started in seconds. None if the mechanism did not start running"""
+    def time(self) -> float:
+        """Elapsed time since mechanism started in seconds. 0.0 if the mechanism did not start running"""
         if self._start_time is None:
             return 0.0
 
-        return time.monotonic() - self._start_time
+        return time.perf_counter() - self._start_time
 
     @property
     def remaining_time(self) -> Optional[float]:
@@ -364,7 +372,7 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
         if self.ami.time_limit == float("+inf"):
             return None
 
-        limit = self.ami.time_limit - (time.monotonic() - self._start_time)
+        limit = self.ami.time_limit - (time.perf_counter() - self._start_time)
         if limit < 0.0:
             return 0.0
 
@@ -377,9 +385,7 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
             return None
 
         relative_step = (
-            (self._step + 1) / self.ami.n_steps
-            if self.ami.n_steps is not None
-            else -1.0
+            self._step / self.ami.n_steps if self.ami.n_steps is not None else -1.0
         )
         relative_time = (
             self.time / self.ami.time_limit if self.ami.time_limit is not None else -1.0
@@ -450,6 +456,10 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
             return True
         return None
 
+    def get_negotiator(self, nid: str) -> Optional["SAONegotiator"]:
+        """Returns the negotiator with the given ID if present in the negotiation"""
+        return self._negotiator_map.get(nid, None)
+
     def remove(self, negotiator: "Negotiator", **kwargs) -> Optional[bool]:
         """Remove the agent from the negotiation.
 
@@ -726,6 +736,8 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
             - There is another function (`run()`) that runs the whole mechanism in blocking mode
 
         """
+        if self._start_time is None or self._start_time < 0:
+            self._start_time = time.perf_counter()
         self.checkpoint_on_step_started()
         state = self.state
         state4history = self.state4history
@@ -752,7 +764,7 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
         if not self._running:
             self._running = True
             self._step = 0
-            self._start_time = time.monotonic()
+            self._start_time = time.perf_counter()
             self._started = True
             if self.on_negotiation_start() is False:
                 self._agreement, self._broken, self._timedout = None, False, False
@@ -780,6 +792,15 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
         step_start = time.perf_counter()
         result = self.round()
         step_time = time.perf_counter() - step_start
+        self._stats["round_times"].append(step_time)
+        if result.times:
+            for k, v in result.times.items():
+                if v is not None:
+                    self._stats["times"][k] += v
+        if result.exceptions:
+            for k, v in result.exceptions.items():
+                if v:
+                    self._stats["exceptions"][k] += v
         state = self.state
         self._error, self._error_details, self._waiting = (
             result.error,
@@ -944,6 +965,10 @@ class Mechanism(NamedObject, EventSource, CheckpointMixin, ABC):
     def history(self):
         return self._history
 
+    @property
+    def stats(self):
+        return self._stats
+
     @property
     def current_step(self):
         return self._step
