@@ -1,4 +1,6 @@
-"""This module defines the interfaces to all negotiation agents (negotiators) in the platform.
+"""
+This module defines the interfaces to all negotiation agents (negotiators)
+in negmas.
 
 """
 import functools
@@ -6,8 +8,6 @@ import math
 import warnings
 from abc import ABC
 from random import sample
-
-import numpy as np
 from typing import (
     TYPE_CHECKING,
     Any,
@@ -20,8 +20,9 @@ from typing import (
     Union,
 )
 
-from negmas.common import *
-from negmas.common import Rational
+import numpy as np
+
+from negmas.common import AgentMechanismInterface, MechanismState, Rational
 from negmas.events import Notifiable, Notification
 from negmas.helpers import get_class
 from negmas.outcomes import Issue
@@ -29,6 +30,7 @@ from negmas.outcomes import Issue
 if TYPE_CHECKING:
     from negmas.outcomes import Outcome
     from negmas.utilities import UtilityValue, UtilityFunction
+    from negmas.situated import Agent
 
 __all__ = [
     "Negotiator",
@@ -398,27 +400,29 @@ class PassThroughNegotiator(Negotiator):
 class Controller(Rational):
     """Controls the behavior of multiple negotiators in multiple negotiations
 
-    The controller class MUST implement any methods of the negotiator class it is controlling with one added
-    argument negotiator_id (str) which represents ID of the negotiator on which the method is being invoked
-    (passed first).
+    The controller class MUST implement any methods of the negotiator class it
+    is controlling with one added argument negotiator_id (str) which represents
+    ID of the negotiator on which the method is being invoked (passed first).
 
-    Controllers for specific classes should inherit from this class and implement whatever methods they want to override
-    on their `PassThroughNegotiator` objects. For example, the SAO module defines `SAOController` that needs only to
-    implement `propose` and `respond` .
+    Controllers for specific classes should inherit from this class and
+    implement whatever methods they want to override on their
+    `PassThroughNegotiator` objects. For example, the SAO module defines
+    `SAOController` that needs only to implement `propose` and `respond` .
 
     Args:
         default_negotiator_type: The negotiator type to use for adding negotiator
                                  if no type is explicitly given.
-        default_negotiator_params: The parameters to use to construct the default
-                                   negotiator type.
+        default_negotiator_params: The parameters to use to construct the
+                                   default negotiator type.
         parent: The parent which can be an `Agent` or another `Controller`
-        auto_kill: If True, negotiators will be killed once their negotiation finishes.
+        auto_kill: If True, negotiators will be killed once their negotiation
+                   finishes.
         name: The controller name
 
     Remarks:
 
-     - Controllers should always call negotiator methods using the `call` method defined in this class. Direct calls may
-       lead to infinite loops
+     - Controllers should always call negotiator methods using the `call`
+       method defined in this class. Direct calls may lead to infinite loops
 
 
     """
@@ -428,10 +432,11 @@ class Controller(Rational):
         default_negotiator_type: Union[str, Type[PassThroughNegotiator]] = None,
         default_negotiator_params: Dict[str, Any] = None,
         parent: Union["Controller", "Agent"] = None,
-        auto_kill: bool = False,
+        auto_kill: bool = True,
         name: str = None,
+        ufun: "UtilityFunction" = None,
     ):
-        super().__init__(name=name)
+        super().__init__(name=name, ufun=ufun)
         self._negotiators: Dict[str, Tuple["PassThroughNegotiator", Any]] = {}
         if default_negotiator_params is None:
             default_negotiator_params = {}
@@ -548,6 +553,7 @@ class Controller(Rational):
             return
         response = negotiator.before_death(cntxt=cntxt)
         if response or force:
+            negotiator._Negotiator__parent = None
             self._negotiators.pop(negotiator_id, None)
 
     def before_join(
