@@ -3,30 +3,18 @@ Implements controllers for the SAO mechanism.
 """
 import itertools
 import random
-from abc import abstractmethod, ABC
+from abc import ABC, abstractmethod
 from collections import defaultdict
+from functools import lru_cache
 
-from typing import (
-    Dict,
-    List,
-    Optional,
-    Tuple,
-    Union,
-)
+from typing import Dict, List, Optional, Tuple, Union
 
-from .common import SAOState, SAOResponse
-from .negotiators import PassThroughSAONegotiator, SAONegotiator
-from ..common import *
+from .common import SAOResponse, SAOState
+from .negotiators import AspirationNegotiator, PassThroughSAONegotiator, SAONegotiator
+from ..common import AgentMechanismInterface, MechanismState
 from ..negotiators import AspirationMixin, Controller
-from ..outcomes import (
-    Outcome,
-    ResponseType,
-    outcome_is_valid,
-)
-from ..utilities import (
-    UtilityFunction,
-    utility_range,
-)
+from ..outcomes import Outcome, ResponseType, outcome_is_valid
+from ..utilities import UtilityFunction, utility_range
 
 __all__ = [
     "SAOController",
@@ -36,24 +24,37 @@ __all__ = [
     "SAOSingleAgreementController",
     "SAOSingleAgreementRandomController",
     "SAOSingleAgreementAspirationController",
+    "SAOMetaNegotiatorController",
 ]
 
 
 class SAOController(Controller):
-    """A controller that can manage multiple negotiators taking full or partial control from them."""
+    """
+    A controller that can manage multiple negotiators taking full or partial control from them.
+
+    Args:
+         default_negotiator_type: Default type to use when creating negotiators using this controller. The default type is
+                                  `PassThroughSAONegotiator` which passes *full control* to the controller.
+         default_negotiator_params: Default paramters to pass to the default controller.
+         auto_kill: Automatically kill the negotiator once its negotiation session is ended.
+         name: Controller name
+         ufun: The ufun of the controller.
+    """
 
     def __init__(
         self,
         default_negotiator_type=PassThroughSAONegotiator,
         default_negotiator_params=None,
-        auto_kill=False,
+        auto_kill=True,
         name=None,
+        ufun=None,
     ):
         super().__init__(
             default_negotiator_type=default_negotiator_type,
             default_negotiator_params=default_negotiator_params,
             auto_kill=auto_kill,
             name=name,
+            ufun=ufun,
         )
 
     def before_join(
@@ -122,13 +123,22 @@ class SAOController(Controller):
         if self._auto_kill:
             self.kill_negotiator(negotiator_id, True)
 
+    def on_negotiation_start(self, negotiator_id: str, state: MechanismState) -> None:
+        pass
+
 
 class SAORandomController(SAOController):
-    """A controller that returns random offers"""
+    """
+    A controller that returns random offers.
 
-    def __init__(self, *args, acceptanbe_prob: float = 0.1, **kwargs):
+    Args:
+        p_acceptance: The probability of accepting an offer.
+
+    """
+
+    def __init__(self, *args, p_acceptance: float = 0.1, **kwargs):
         super().__init__(*args, **kwargs)
-        self._acceptance_prob = acceptanbe_prob
+        self._p_acceptance = p_acceptance
 
     def propose(self, negotiator_id: str, state: MechanismState) -> Optional["Outcome"]:
 
@@ -145,13 +155,23 @@ class SAORandomController(SAOController):
         negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
         if negotiator is None:
             raise ValueError(f"Unknown negotiator {negotiator_id}")
-        if random.random() > self._acceptance_prob:
+        if random.random() > self._p_acceptance:
             return ResponseType.ACCEPT_OFFER
         return ResponseType.REJECT_OFFER
 
 
 class SAOSyncController(SAOController):
-    """A controller that can manage multiple negotiators synchronously"""
+    """
+    A controller that can manage multiple negotiators synchronously.
+
+    Remarks:
+        - The controller waits for an offer from each one of its negotiators before deciding what to do.
+        - Loops may happen if multiple controllers of this type negotiate with each other. For example controller A
+          is negotiating with B, C, while B is also negotiating with C. These loops are broken by the `SAOMechanism`
+          by **forcing** some controllers to respond before they have all of the offers. In this case, `counter_all`
+          will receive offers from one or more negotiators but not all of them.
+
+    """
 
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
@@ -174,6 +194,8 @@ class SAOSyncController(SAOController):
         # if some proposal was there, delete it to force the controller to get a new one
         if proposal is not None:
             self.proposals[negotiator_id] = None
+        # if the proposal that was there was None, just offer the best offer
+        proposal = self.first_offer(negotiator_id)
         return proposal
 
     def respond(
@@ -190,11 +212,12 @@ class SAOSyncController(SAOController):
         # set the saved offer for this negotiator
         self.offers[negotiator_id] = offer
         self.offer_states[negotiator_id] = state
-
+        n_negotiators = len(self.active_negotiators)
         # if we got all the offers or waited long enough, counter all the offers so-far
-        if len(self.offers) == len(self.negotiators) or self.n_waits[
-            negotiator_id
-        ] >= len(self.negotiators):
+        if (
+            len(self.offers) == n_negotiators
+            or self.n_waits[negotiator_id] >= n_negotiators
+        ):
             responses = self.counter_all(offers=self.offers, states=self.offer_states)
             for nid in responses.keys():
                 # register the responses for next time for all other negotiators
@@ -235,6 +258,7 @@ class SAOSyncController(SAOController):
             )
         )
 
+    @lru_cache(maxsize=100)
     def first_offer(self, negotiator_id: str) -> Optional["Outcome"]:
         """
         Finds the first offer for this given negotiator. By default it will be the best offer
@@ -255,19 +279,34 @@ class SAOSyncController(SAOController):
         if negotiator.ami is None:
             return None
         # if the controller has a ufun, use it otherwise use the negotiator ufun
-        if hasattr(self, "ufun"):
+        if self.ufun is not None:
             _, _, _, best = utility_range(
                 self.ufun, issues=negotiator.ami.issues, return_outcomes=True
             )
-        else:
+        elif negotiator.ufun is not None:
             _, _, _, best = utility_range(
                 negotiator.ufun, issues=negotiator.ami.issues, return_outcomes=True
             )
+        else:
+            best = None
         return best
 
 
 class SAORandomSyncController(SAOSyncController):
-    """A sync controller that returns random offers"""
+    """
+    A sync controller that returns random offers. (See `SAOSyncController` ).
+
+    Args:
+        p_acceptance: The probability that an offer will be accepted
+        p_rejection: The probability that an offer will be rejected
+        p_ending: The probability of ending the negotiation at any negotiation round.
+
+    Remarks:
+        - If probability of acceptance, rejection and ending sum to less than 1.0, the agent will return NO_RESPONSE
+          with the remaining probability. Depending on the settings of the `SAOMechanism` this may be treated as
+          ending the negotiation.
+
+    """
 
     def __init__(
         self, *args, p_acceptance=0.15, p_rejection=0.85, p_ending=0.0, **kwargs
@@ -635,7 +674,7 @@ class SAOMetaNegotiatorController(SAOController):
     Controls multiple negotiations using a single `meta` negotiator.
 
     Args:
-        - meta_negotiator: The negotiator used for controlling all negotiations
+        - meta_negotiator: The negotiator used for controlling all negotiations.
 
     Remarks:
 
@@ -655,7 +694,9 @@ class SAOMetaNegotiatorController(SAOController):
     def __init__(self, *args, meta_negotiator: SAONegotiator = None, **kwargs):
         super().__init__(*args, **kwargs)
         if meta_negotiator is None:
-            meta_negotiator = AspirationNegotiator(name=f"{self.name}-negotiator")
+            meta_negotiator = AspirationNegotiator(
+                name=f"{self.name}-negotiator", ufun=kwargs.get("ufun", None)
+            )
         self.meta_negotiator = meta_negotiator
 
     def propose(self, negotiator_id: str, state: MechanismState) -> Optional["Outcome"]:
@@ -678,6 +719,14 @@ class SAOMetaNegotiatorController(SAOController):
 
 
 class SAOSingleAgreementRandomController(SAOSingleAgreementController):
+    """
+    A single agreement controller that uses a random negotiation strategy.
+
+    Args:
+        p_acceptance: The probability that an offer is accepted
+
+    """
+
     def is_acceptable(self, offer: "Outcome", source: str, state: SAOState) -> bool:
         return random.random() > self.p_acceptance
 
@@ -720,14 +769,12 @@ class SAOSingleAgreementAspirationController(
     def __init__(
         self,
         *args,
-        ufun: UtilityFunction,
         max_aspiration: float = 1.0,
         aspiration_type: Union[str, int, float] = "boulware",
         **kwargs,
     ):
         super().__init__(*args, **kwargs)
         AspirationMixin.aspiration_init(self, max_aspiration, aspiration_type)
-        self.ufun = ufun
 
     def after_join(self, negotiator_id, ami, state, *, ufun=None, role="agent"):
         if self.ufun is not None:
