@@ -12,6 +12,7 @@ import random
 import time
 import traceback
 import warnings
+from collections import defaultdict
 from dataclasses import dataclass, field
 from multiprocessing import cpu_count
 from os import PathLike
@@ -36,6 +37,7 @@ from scipy.stats import ks_2samp, ttest_ind
 from typing_extensions import Protocol
 
 from negmas.helpers import (
+    exception2str,
     add_records,
     dump,
     get_class,
@@ -45,6 +47,7 @@ from negmas.helpers import (
     load,
     unique_name,
 )
+from negmas.java import to_flat_dict
 
 from .situated import Agent, World, save_stats
 
@@ -71,7 +74,7 @@ except:
     pass
 
 
-def _hash(*args):
+def _hash(*args) -> str:
     """Generates a unique ID given any inputs"""
     return hashlib.sha1(
         ("h" + "".join([str(_) for _ in args])).encode("utf-8")
@@ -155,6 +158,9 @@ class ConfigAssigner(Protocol):
         fair: bool = True,
         competitors: Sequence[Union[str, Type[Agent]]] = (),
         params: Sequence[Dict[str, Any]] = (),
+        dynamic_non_competitors: Sequence[Union[str, Type[Agent]]] = None,
+        dynamic_non_competitor_params: Sequence[Dict[str, Any]] = None,
+        exclude_competitors_from_reassignment: bool = True,
     ) -> List[List[Dict[str, Any]]]:
         ...
 
@@ -189,14 +195,24 @@ class WorldSetRunStats:
     """Actually executed number of steps for each world"""
     execution_time: int
     """Total execution time of each world"""
-    n_agent_exceptions: int = 0
-    """Exceptions in agent code"""
-    n_negotiation_exceptions: int = 0
-    """Exceptions in negotiations"""
-    n_other_exceptions: int = 0
-    """Exceptions when running the tournament"""
-    n_simulation_exceptions: int = 0
-    """Exceptions when running the tournament"""
+    simulation_exceptions: Dict[int, List[str]] = field(default_factory=dict)
+    """Exceptions thrown by the simulator (not including mechanism creation and contract exceptions)"""
+    contract_exceptions: Dict[int, List[str]] = field(default_factory=dict)
+    """Exceptions thrown by the simulator during contract execution"""
+    mechanism_exceptions: Dict[int, List[str]] = field(default_factory=dict)
+    """Exceptions thrown by the simulator during mechanism creation or execution"""
+    other_exceptions: List[str] = field(default_factory=list)
+    """Exceptions raised by tournament running code itself not any world"""
+    agent_exceptions: Dict[str, List[Tuple[int, str]]] = field(
+        default_factory=lambda: defaultdict(list)
+    )
+    """All exceptions thrown per agent (not including negotiator exceptions)"""
+    negotiator_exceptions: Dict[str, List[Tuple[int, str]]] = field(
+        default_factory=lambda: defaultdict(list)
+    )
+    """All exceptions thown by negotiators of an agent"""
+    agent_times: Dict[str, float] = field(default_factory=lambda: defaultdict(float))
+    """Total execution time per agent"""
 
 
 @dataclass
@@ -222,7 +238,7 @@ class TournamentResults:
     path: str = None
     """Path at which tournament results are stored"""
     world_stats: pd.DataFrame = None
-    """Some satistics about each world run"""
+    """Some statistics about each world run"""
     params: Dict[str, Any] = None
     """Parameters of the tournament"""
 
@@ -242,7 +258,7 @@ class TournamentResults:
 
 def run_world(
     world_params: dict, dry_run: bool = False, save_world_stats: bool = True
-) -> Tuple[str, WorldRunResults, WorldSetRunStats]:
+) -> Tuple[str, Optional[WorldRunResults], WorldSetRunStats]:
     """Runs a world and returns stats. This function is designed to be used with distributed systems like dask.
 
     Args:
@@ -296,7 +312,7 @@ def run_world(
 
 def run_worlds(
     worlds_params: List[dict], dry_run: bool = False, save_world_stats: bool = True
-) -> Tuple[str, WorldRunResults, WorldSetRunStats]:
+) -> Tuple[str, Optional[WorldRunResults], WorldSetRunStats]:
     """Runs a set of worlds and returns stats. This function is designed to be used with distributed systems like dask.
 
     Args:
@@ -320,7 +336,9 @@ def run_worlds(
     if len(worlds_params) < 1:
         return (
             _hash(worlds_params),
-            WorldRunResults(world_names=[""], log_file_names=[""]),
+            # WorldRunResults(world_names=[""], log_file_names=[""]),
+            None,
+            None,
         )
     world_generator, score_calculator = None, None
     for world_params in worlds_params:
@@ -364,7 +382,9 @@ def _run_worlds(
     world_progress_callback: Callable[[Optional[World]], None] = None,
     dry_run: bool = False,
     save_world_stats: bool = True,
-) -> Tuple[str, WorldRunResults, WorldSetRunStats]:
+    override_ran_worlds: bool = False,
+    save_progress_every: int = 1,
+) -> Tuple[str, Optional[WorldRunResults], WorldSetRunStats]:
     """Runs a set of worlds (generated from a world generator) and returns stats
 
     Args:
@@ -374,6 +394,8 @@ def _run_worlds(
         world_progress_callback: world progress callback
         dry_run: If true, the world is not run. Its config is saved instead.
         save_world_stats: If true, saves individual world stats
+        override_ran_worlds: If true, run the worlds even if they are already ran before.
+        save_progress_every: If true, progress will be saved every this number of steps.
 
     Returns:
         A tuple with the following components in order:
@@ -392,11 +414,16 @@ def _run_worlds(
             - __dir_name: directory to store the world stats
             - others: values of all other keys are passed to the world generator as kwargs
 
+        - The system knows that a world is already ran using `is_already_run` which checks that
+          the folder to store the results of the world exists and contains a file called stats.json
+
     """
     worlds, dir_names = [], []
     scoring_context = {}
     run_id = _hash(worlds_params)
     video_savers, video_saver_params_list, save_videos = [], [], []
+    scores: Optional[WorldRunResults] = None
+
     for world_params in worlds_params:
         world_params = world_params.copy()
         dir_name = world_params["__dir_name"]
@@ -415,22 +442,46 @@ def _run_worlds(
             world.save_config(dir_name)
             continue
     try:
-        for world, dir_name, save_video, video_saver, video_saver_params in zip(
-            worlds, dir_names, save_videos, video_savers, video_saver_params_list
+        for (
+            world,
+            world_params_,
+            dir_name,
+            save_video,
+            video_saver,
+            video_saver_params,
+        ) in zip(
+            worlds,
+            world_params,
+            dir_names,
+            save_videos,
+            video_savers,
+            video_saver_params_list,
         ):
-            if world_progress_callback is None:
-                world.run()
-            else:
-                _start_time = time.monotonic()
-                for _ in range(world.n_steps):
-                    if (
-                        world.time_limit is not None
-                        and (time.monotonic() - _start_time) >= world.time_limit
-                    ):
-                        break
-                    if not world.step():
-                        break
-                    world_progress_callback(world)
+            _start_time = time.perf_counter()
+            for _ in range(world.n_steps):
+                if not world.step():
+                    break
+                if _ % save_progress_every == 0:
+                    save_stats(world, world.log_folder, params=world_params_)
+                    # TODO reorganize the code so that the worlds are run in parallel when there are multiple of them
+                    if not dry_run:
+                        scores_ = to_flat_dict(
+                            score_calculator(worlds, scoring_context, False)
+                        )
+                        scores_["n_steps"] = world.n_steps
+                        scores_["step"] = world.current_step
+                        scores_["relative_time"] = world.relative_time
+                        scores_["time_limit"] = world.time_limit
+                        scores_["time"] = world.time
+                        dump(
+                            to_flat_dict(scores_),
+                            Path(world.log_folder) / "_current_scores.json",
+                            sort_keys=True,
+                        )
+                    if world_progress_callback:
+                        world_progress_callback(world)
+                if world.time >= world.time_limit:
+                    break
             if save_world_stats:
                 save_stats(world=world, log_dir=dir_name)
             if save_video:
@@ -441,15 +492,51 @@ def _run_worlds(
                 video_saver(world, **video_saver_params)
 
         scores = score_calculator(worlds, scoring_context, dry_run)
+        agent_exceptions: Dict[str, List[Tuple[int, str]]] = defaultdict(list)
+        for w in worlds:
+            for aid, v in w.agent_exceptions.items():
+                if v:
+                    agent_exceptions[w.agents[aid].type_name] += v
+        negotiator_exceptions: Dict[str, List[Tuple[int, str]]] = defaultdict(list)
+        for w in worlds:
+            for aid, v in w.negotiator_exceptions.items():
+                if v:
+                    negotiator_exceptions[w.agents[aid].type_name] += v
+
+        simulation_exceptions: Dict[int, List[str]] = defaultdict(list)
+        for w in worlds:
+            for k, l in w.simulation_exceptions.items():
+                if l:
+                    simulation_exceptions[k] += l
+        mechanism_exceptions: Dict[int, List[str]] = defaultdict(list)
+        for w in worlds:
+            for k, l in w.mechanism_exceptions.items():
+                if l:
+                    mechanism_exceptions[k] += l
+        contract_exceptions: Dict[int, List[str]] = defaultdict(list)
+        for w in worlds:
+            for k, l in w.contract_exceptions.items():
+                if l:
+                    contract_exceptions[k] += l
+
+        agent_times: Dict[str, float] = defaultdict(float)
+        for w in worlds:
+            for aid, _ in w.times.items():
+                if _:
+                    agent_times[w.agents[aid].type_name] += _
+
         world_stats = WorldSetRunStats(
             name=";".join(_.name for _ in worlds),
             planned_n_steps=sum(_.n_steps for _ in worlds),
             executed_n_steps=sum(_.current_step for _ in worlds),
             execution_time=sum(_.frozen_time for _ in worlds),
-            n_negotiation_exceptions=sum(_.n_negotiation_exceptions for _ in worlds),
-            n_simulation_exceptions=sum(_.n_simulation_exceptions for _ in worlds),
-            n_other_exceptions=0,
-            n_agent_exceptions=sum(_.n_agent_exceptions for _ in worlds),
+            agent_exceptions=agent_exceptions,
+            negotiator_exceptions=negotiator_exceptions,
+            simulation_exceptions=simulation_exceptions,
+            contract_exceptions=contract_exceptions,
+            mechanism_exceptions=mechanism_exceptions,
+            agent_times=agent_times,
+            other_exceptions=[],
         )
     except Exception as e:
         scores = None
@@ -460,19 +547,17 @@ def _run_worlds(
             planned_n_steps=sum(_.n_steps for _ in worlds),
             executed_n_steps=sum(_.current_step for _ in worlds),
             execution_time=sum(_.frozen_time for _ in worlds),
-            n_negotiation_exceptions=sum(_.n_negotiation_exceptions for _ in worlds),
-            n_simulation_exceptions=sum(_.n_simulation_exceptions for _ in worlds),
-            n_other_exceptions=1,
-            n_agent_exceptions=sum(_.n_agent_exceptions for _ in worlds),
+            other_exceptions=[exception2str()],
         )
     return run_id, scores, world_stats
 
 
 def process_world_run(
     run_id: str,
-    results: WorldRunResults,
+    results: Optional[WorldRunResults],
     tournament_name: str,
-    save_world_stats: bool = True,
+    world_stats: Optional[WorldSetRunStats],
+    # save_world_stats: bool = True,
 ) -> List[Dict[str, Any]]:
     """
     Generates a data-frame with the results of this world run
@@ -481,7 +566,6 @@ def process_world_run(
         run_id: The ID of this run (should be unique per tournament)
         results: Results of the world run
         tournament_name: tournament name
-        save_world_stats: It True, it will be assumed that world stats are saved
 
     Returns:
 
@@ -493,8 +577,8 @@ def process_world_run(
     log_files, world_names_ = results.log_file_names, results.world_names
     for world_name_, log_file in zip(world_names_, log_files):
         if (
-            save_world_stats
-            and log_file is not None
+            # save_world_stats
+            log_file is not None
             and pathlib.Path(log_file).exists()
         ):
             with open(log_file, "a") as f:
@@ -508,6 +592,9 @@ def process_world_run(
         for log_file_name in log_files
     )
     base_folder = str(pathlib.Path(log_files[0]).parent) if log_files[0] else ""
+    total_time = sum(world_stats.agent_times.values())
+    if total_time < 1e-6:
+        total_time = 1
     for name_, type_, score in zip(results.names, results.types, results.scores):
         d = {
             "agent_name": name_,
@@ -518,6 +605,9 @@ def process_world_run(
             "stats_folders": stat_folders,
             "base_stats_folder": base_folder,
             "run_id": run_id,
+            "time": world_stats.agent_times.get(type_, 0) / total_time,
+            "exceptions": world_stats.agent_exceptions.get(type_, []),
+            "negotiator_exceptions": world_stats.negotiator_exceptions.get(type_, []),
         }
         scores.append(d)
     return scores
@@ -539,6 +629,7 @@ def _run_dask(
     world_stats_file,
     run_ids,
     print_exceptions,
+    override_already_ran=False,
 ) -> None:
     """Runs the tournament on dask"""
 
@@ -569,6 +660,7 @@ def _run_dask(
                 None,
                 dry_run,
                 save_world_stats,
+                override_already_ran,
             )
         )
     print(f"Submitted all processes to DASK ({len(future_results)})")
@@ -586,10 +678,11 @@ def _run_dask(
                     run_id,
                     score_,
                     tournament_name=name,
-                    save_world_stats=save_world_stats,
+                    world_stats=world_stats_,
+                    # save_world_stats=save_world_stats,
                 ),
-                add_records(world_stats_file, [vars(world_stats_)]),
             )
+            add_records(world_stats_file, [vars(world_stats_)])
             if verbose:
                 _duration = time.perf_counter() - _strt
                 print(
@@ -655,6 +748,9 @@ def tournament(
     world_progress_callback: Callable[[Optional[World]], None] = None,
     non_competitors: Optional[Tuple[Union[str, Any]]] = None,
     non_competitor_params: Optional[Tuple[Dict[str, Any]]] = None,
+    dynamic_non_competitors: Optional[Tuple[Union[str, Any]]] = None,
+    dynamic_non_competitor_params: Optional[Tuple[Dict[str, Any]]] = None,
+    exclude_competitors_from_reassignment: bool = True,
     name: str = None,
     verbose: bool = False,
     configs_only: bool = False,
@@ -720,9 +816,12 @@ def tournament(
                                  and parallel evaluation and should be used with cautious).
         tournament_progress_callback: A function to be called with `WorldRunResults` after each world finished
                                       processing
-        non_competitors: A list of agent types that will not be competing in the sabotage competition but will exist
-                         in the world
+        non_competitors: A list of agent types that will not be competing but will still exist in the world.
         non_competitor_params: paramters of non competitor agents
+        dynamic_non_competitors: A list of non-competing agents that are assigned to the simulation dynamically during
+                                 the creation of the final assignment instead when the configuration is created
+        dynamic_non_competitor_params: paramters of dynamic non competitor agents
+        exclude_competitors_from_reassignment: If true, competitors are excluded from the dyanamic non-competitors
         verbose: Verbosity
         configs_only: If true, a config file for each
         compact: If true, compact logs will be created and effort will be made to reduce the memory footprint
@@ -794,6 +893,9 @@ def tournament(
             scheduler_port=scheduler_port,
             non_competitors=non_competitors,
             non_competitor_params=non_competitor_params,
+            dynamic_non_competitors=dynamic_non_competitors,
+            dynamic_non_competitor_params=dynamic_non_competitor_params,
+            exclude_competitors_from_reassignment=exclude_competitors_from_reassignment,
             name=stage_name,
             verbose=verbose,
             compact=compact,
@@ -879,6 +981,16 @@ def _path(path: Union[str, PathLike]) -> Path:
     return pathlib.Path(path).absolute()
 
 
+def is_already_run(world_params) -> bool:
+    return False
+    # dir_name = pathlib.Path(world_params["__dir_name"])
+    # if not dir_name.exists():
+    #     return False
+    # if (dir_name / "stats.json").exists():
+    #     return True
+    # return False
+
+
 def run_tournament(
     tournament_path: Union[str, PathLike],
     world_generator: WorldGenerator = None,
@@ -896,6 +1008,7 @@ def run_tournament(
     verbose: bool = False,
     compact: bool = None,
     print_exceptions: bool = True,
+    override_ran_worlds: bool = False,
 ) -> None:
     """
     Runs a tournament
@@ -922,6 +1035,7 @@ def run_tournament(
         verbose: Verbosity
         compact: If true, compact logs will be created and effort will be made to reduce the memory footprint
         print_exceptions: If true, exceptions encountered during world simulation will be printed to stdout
+        override_ran_worlds: If true worlds that are already ran will be ran again
 
     """
     tournament_path = _path(tournament_path)
@@ -943,6 +1057,7 @@ def run_tournament(
         compact = params.get("compact", False)
 
     assigned = load(tournament_path / "assigned_configs.pickle")
+    random.shuffle(assigned)
     n_world_configs = len(assigned)
 
     if verbose:
@@ -996,13 +1111,14 @@ def run_tournament(
                     score_calculator=score_calculator,
                     dry_run=False,
                     save_world_stats=True,
+                    override_ran_worlds=override_ran_worlds,
                 )
                 if tournament_progress_callback is not None:
                     tournament_progress_callback(score_, i, n_world_configs)
                 add_records(
                     scores_file,
                     process_world_run(
-                        run_id, score_, tournament_name=name, save_world_stats=True
+                        run_id, score_, tournament_name=name, world_stats=world_stats_,
                     ),
                 )
                 add_records(world_stats_file, [vars(world_stats_)])
@@ -1044,8 +1160,10 @@ def run_tournament(
                     world_progress_callback,
                     False,
                     True,
+                    override_ran_worlds,
                 )
             )
+        result_received = set()
         if verbose:
             print(
                 f"Submitted all processes ({len(future_results)} of {len(assigned)})",
@@ -1060,6 +1178,8 @@ def run_tournament(
         for i, future in enumerate(
             futures.as_completed(future_results, timeout=total_timeout)
         ):
+            if total_timeout is not None and time.perf_counter() - strt > total_timeout:
+                break
             try:
                 run_id, score_, world_stats_ = future.result()
                 if tournament_progress_callback is not None:
@@ -1070,10 +1190,12 @@ def run_tournament(
                         run_id,
                         score_,
                         tournament_name=name,
-                        save_world_stats=not compact,
+                        world_stats=world_stats_,
+                        # save_world_stats=not compact,
                     ),
-                    add_records(world_stats_file, [vars(world_stats_)]),
                 )
+                add_records(world_stats_file, [vars(world_stats_)]),
+                result_received.add(run_id)
                 if verbose:
                     _duration = time.perf_counter() - _strt
                     print(
@@ -1116,6 +1238,7 @@ def run_tournament(
             world_stats_file,
             run_ids,
             print_exceptions,
+            override_ran_worlds,
         )
     if verbose:
         print(f"Tournament completed successfully")
@@ -1144,17 +1267,20 @@ def create_tournament(
     scheduler_port: Optional[str] = None,
     non_competitors: Optional[Tuple[Union[str, Any]]] = None,
     non_competitor_params: Optional[Tuple[Dict[str, Any]]] = None,
+    dynamic_non_competitors: Optional[Tuple[Union[str, Any]]] = None,
+    dynamic_non_competitor_params: Optional[Tuple[Dict[str, Any]]] = None,
+    exclude_competitors_from_reassignment: bool = True,
     name: str = None,
     verbose: bool = False,
     compact: bool = False,
-    save_video_fraction: float = 0.001,
-    forced_logs_fraction: float = 0.001,
+    save_video_fraction: float = 0.0,
+    forced_logs_fraction: float = 0.0,
     video_params=None,
     video_saver=None,
     **kwargs,
 ) -> PathLike:
     """
-    Runs a tournament
+    Creates a tournament
 
     Args:
 
@@ -1200,9 +1326,13 @@ def create_tournament(
                      to use half of the CPU in the machine). By defaults parallel uses all CPUs in the machine
         scheduler_port: Port of the dask scheduler if parallelism is dask, dist, or distributed
         scheduler_ip:   IP Address of the dask scheduler if parallelism is dask, dist, or distributed
-        non_competitors: A list of agent types that will not be competing in the sabotage competition but will exist
-                         in the world
+        non_competitors: A list of agent types that will not be competing but will still exist in the world.
         non_competitor_params: paramters of non competitor agents
+        dynamic_non_competitors: A list of non-competing agents that are assigned to the simulation dynamically during
+                                 the creation of the final assignment instead when the configuration is created
+        dynamic_non_competitor_params: paramters of dynamic non competitor agents
+        exclude_competitors_from_reassignment: If true, copmetitors are not included in the reassignment even
+                                                          if they exist in `dynamic_non_competitors`
         verbose: Verbosity
         compact: If true, compact logs will be created and effort will be made to reduce the memory footprint
         save_video_fraction: The fraction of simulations for which to save videos
@@ -1348,6 +1478,9 @@ def create_tournament(
                         n_agents_per_competitor=n_agents_per_competitor,
                         competitors=effective_competitors,
                         params=effective_params,
+                        dynamic_non_competitors=dynamic_non_competitors,
+                        dynamic_non_competitor_params=dynamic_non_competitor_params,
+                        exclude_competitors_from_reassignment=exclude_competitors_from_reassignment,
                     )
                     for c in myconfigs
                 ]
@@ -1421,7 +1554,7 @@ def create_tournament(
             config["world_params"].update(
                 {"log_file_name": str("log.txt"), "log_folder": str(dir_name)}
             )
-    if forced_logs_fraction > 0.0:
+    if forced_logs_fraction > 1e-5:
         n_logged = max(1, int(len(assigned) * forced_logs_fraction))
         for cs in assigned[:n_logged]:
             for _ in cs:
@@ -1439,7 +1572,7 @@ def create_tournament(
                         }
                     )
 
-    if save_video_fraction > 0.0:
+    if save_video_fraction > 1e-5:
         n_videos = max(1, int(len(assigned) * forced_logs_fraction))
         for cs in assigned[:n_videos]:
             for _ in cs:
@@ -1732,24 +1865,24 @@ def combine_tournament_stats(
         src = _path(src)
         for filename in src.glob("**/stats.json"):
             # try:
-            p = load(filename)
-            # removing default factory managers from the dict because the balances/storages are mixed
-            # @todo unmix balances/scores of default fms.
-            # p = dict(
-            #     zip(
-            #         [c for c in p.keys() if "_df_" not in c],
-            #         [p[c] for c in p.keys() if "_df_" not in c],
-            #     )
-            # )
-            p = pd.DataFrame.from_dict(p)
-            p = p.loc[:, [c for c in p.columns if World.is_basic_stat(c)]]
-            p["step"] = list(range(len(p)))
-            p["world"] = filename.parent.name
-            p["path"] = filename.parent.parent
-            stats.append(p)
-            # if verbose:
-            #    print(f"Read: {str(filename)}")
-            #        print(f"FAILED {str(filename)}")
+            data = load(filename)
+            if data is None or len(data) == 0:
+                continue
+            try:
+                data = pd.DataFrame.from_dict(data)
+            except:
+                # adjust lengths. Some columns are longer than others
+                min_len = min(len(_) for _ in data.values())
+                for k, v in data.items():
+                    if len(v) == min_len:
+                        continue
+                    data[k] = data[k][:min_len]
+                data = pd.DataFrame.from_dict(data)
+            data = data.loc[:, [c for c in data.columns if World.is_basic_stat(c)]]
+            data["step"] = list(range(len(data)))
+            data["world"] = filename.parent.name
+            data["path"] = filename.parent.parent
+            stats.append(data)
     if len(stats) < 1:
         if verbose:
             print("No stats found")
